from django.core.exceptions import ValidationError

from ..caluma_form.validators import DocumentValidator
from . import models


class WorkItemValidator:
    def _validate_task_simple(self, task, case, document, user, **kwargs):
        pass

    def _validate_task_complete_workflow_form(
        self, task, case, document, user, **kwargs
    ):
        self._validate_task_complete_task_form(
            task, case, case.document, user, **kwargs
        )

    def _validate_task_complete_task_form(
        self,
        task,
        case,
        document,
        user,
        validation_context=None,
        data_source_context=None,
        **kwargs,
    ):
        DocumentValidator().validate(
            document,
            user,
            validation_context=validation_context,
            data_source_context=data_source_context,
            **kwargs,
        )

    def validate(
        self,
        *,
        status,
        child_case,
        case,
        task,
        document,
        user,
        validation_context=None,
        data_source_context=None,
        **kwargs,
    ):
        if case.status != models.Case.STATUS_RUNNING:
            raise ValidationError("Only work items of running cases can be completed.")

        if status != models.WorkItem.STATUS_READY:
            raise ValidationError("Only ready work items can be completed.")

        if child_case:
            if child_case.status == models.Case.STATUS_RUNNING:
                raise ValidationError(
                    "Work item can only be completed when child case is in a finish state."
                )

        getattr(self, f"_validate_task_{task.type}")(
            task,
            case,
            document,
            user,
            validation_context=validation_context,
            data_source_context=data_source_context,
            **kwargs,
        )



from io import StringIO

from django.conf import settings
from django.core import management
from watchman.decorators import check

from caluma.caluma_form import storage_clients


@check
def _check_pending_migrations(db_name):
    """Check database for pending migrations.

    Returns JSON mapping if no pending migrations, otherwise raises Exception.
    @check django-watchman decorator catches and handles exceptions.
    """
    # check for unapplied migrations
    out = StringIO()
    management.call_command("showmigrations", "--plan", stdout=out)
    plan = out.getvalue()
    if "[ ]" in plan:
        raise Exception("Database has unapplied migrations (migrate).")

    return {db_name: {"ok": True}}


@check
def _check_media_storage_service():
    """Check media storage service connectivity."""
    return {
        "ok": storage_clients.client.client.bucket_exists(
            settings.MINIO_STORAGE_MEDIA_BUCKET_NAME
        )
    }


def check_media_storage_service():
    """Check media storage service connectivity."""
    # Check if media storage service is available
    if not storage_clients.client:
        return {"media storage service (not configured)": {"ok": True}}

    return {"media storage service": _check_media_storage_service()}


def check_migrations():
    """Check available databases for unapplied migrations."""
    databases = sorted(settings.DATABASES)
    checked_databases = [_check_pending_migrations(db_name) for db_name in databases]

    return {"database migrations": checked_databases}



import functools
import hashlib

import requests
from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured
from django.http.response import HttpResponse
from django.utils.encoding import force_bytes, smart_str
from django.utils.module_loading import import_string
from graphene.validation import DisableIntrospection, depth_limit_validator
from graphene_django.views import GraphQLView, HttpError
from rest_framework.authentication import get_authorization_header

from caluma.caluma_user import models


class HttpResponseUnauthorized(HttpResponse):
    status_code = 401


custom_validation_rules = []
if settings.DISABLE_INTROSPECTION:  # pragma: no cover
    custom_validation_rules.append(DisableIntrospection)

if settings.QUERY_DEPTH_LIMIT:  # pragma: no cover
    custom_validation_rules.append(
        depth_limit_validator(max_depth=settings.QUERY_DEPTH_LIMIT)
    )


class AuthenticationGraphQLView(GraphQLView):
    if custom_validation_rules:  # pragma: no cover
        validation_rules = tuple(custom_validation_rules)

    def get_bearer_token(self, request):
        auth = get_authorization_header(request).split()
        header_prefix = "Bearer"

        if not auth:
            return None

        if smart_str(auth[0].lower()) != header_prefix.lower():
            raise HttpError(HttpResponseUnauthorized("No Bearer Authorization header"))

        if len(auth) == 1:
            msg = "Invalid Authorization header. No credentials provided"
            raise HttpError(HttpResponseUnauthorized(msg))
        elif len(auth) > 2:
            msg = (
                "Invalid Authorization header. Credentials string should "
                "not contain spaces."
            )
            raise HttpError(HttpResponseUnauthorized(msg))

        return auth[1]

    def get_userinfo(self, token):
        response = requests.get(
            settings.OIDC_USERINFO_ENDPOINT,
            verify=settings.OIDC_VERIFY_SSL,
            headers={"Authorization": f"Bearer {smart_str(token)}"},
        )
        response.raise_for_status()
        return response.json()

    def _oidc_user(self, *args, **kwargs):
        factory = import_string(settings.CALUMA_OIDC_USER_FACTORY)
        return factory(*args, **kwargs)

    def get_user(self, request):
        token = self.get_bearer_token(request)

        if token is None:
            return models.AnonymousUser()

        if not settings.OIDC_USERINFO_ENDPOINT:
            raise ImproperlyConfigured(
                'Token was provided, but "OIDC_USERINFO_ENDPOINT" is not set.'
            )

        userinfo_method = functools.partial(self.get_userinfo, token=token)
        # token might be too long for key, so we use hash sum instead.
        hashsum_token = hashlib.sha256(force_bytes(token)).hexdigest()

        try:
            claims = cache.get_or_set(
                f"authentication.userinfo.{hashsum_token}",
                userinfo_method,
                timeout=settings.OIDC_BEARER_TOKEN_REVALIDATION_TIME,
            )
        except requests.HTTPError as e:
            # convert request error to django http error
            response = HttpResponse()
            response.status_code = e.response.status_code
            raise HttpError(response, message=str(e))

        return self._oidc_user(token=token, claims=claims)

    def dispatch(self, request, *args, **kwargs):
        try:
            request.user = self.get_user(request)
            return super().dispatch(request, *args, **kwargs)
        except HttpError as e:
            response = e.response
            response["Content-Type"] = "application/json"
            response.content = self.json_encode(
                request, {"errors": [self.format_error(e)]}
            )
            return response



import itertools

from django.db.models.signals import (
    m2m_changed,
    post_delete,
    post_init,
    post_save,
    pre_delete,
    pre_save,
)
from django.dispatch import receiver
from simple_history.signals import pre_create_historical_record

from caluma.caluma_core.events import filter_events
from caluma.utils import disable_raw

from . import models
from .utils import (
    recalculate_answers_from_document,
    update_calc_dependents,
    update_or_create_calc_answer,
)


@receiver(pre_create_historical_record, sender=models.HistoricalAnswer)
def set_historical_answer_type(sender, **kwargs):
    """Receiver for historical answer pre_save signal setting question type."""
    historical_answer = kwargs["history_instance"]

    try:
        # can take the current question, assuming the question type is not changed in a datarace
        historical_answer.history_question_type = historical_answer.question.type
    except models.Question.DoesNotExist:
        # This happens when deleting a question with a default_answer. As the Answer
        # has to exist for this to happen, we can safely assume to already have a
        # historical record of it. So we're taking the question type from the previous
        # revision.
        previous = historical_answer.get_previous_by_history_date()
        historical_answer.history_question_type = previous.history_question_type


@receiver(post_delete, sender=models.Question)
@filter_events(lambda instance: instance.default_answer is not None)
def cleanup_default_answer(sender, instance, **kwargs):
    """Ensure default_answers are cleanedup."""
    instance.default_answer.delete()


@receiver(post_init, sender=models.Document)
@filter_events(lambda instance: instance.family_id is None)
def set_document_family(sender, instance, **kwargs):
    """
    Ensure document has the family pointer set.

    This sets the document's family if not overruled or set already.
    The family will be in the corresponding mutations when a tree
    structure is created or restructured.
    """

    # can't use instance.set_family() here as the instance isn't
    # in DB yet, causing integrity errors
    instance.family = instance


# Update calc dependents on pre_save
#
# Every question that is referenced in a `calcExpression` will memoize the
# referencing calculated question. This list of calculation dependents is
# mutated whenever a calculated question is created, updated or deleted.


@receiver(pre_save, sender=models.Question)
@disable_raw
@filter_events(lambda instance: instance.type == models.Question.TYPE_CALCULATED_FLOAT)
def save_calc_dependents(sender, instance, **kwargs):
    original = models.Question.objects.filter(pk=instance.pk).first()
    if not original:
        update_calc_dependents(
            instance.slug, old_expr="false", new_expr=instance.calc_expression
        )

    elif original.calc_expression != instance.calc_expression:
        update_calc_dependents(
            instance.slug,
            old_expr=original.calc_expression,
            new_expr=instance.calc_expression,
        )


@receiver(pre_delete, sender=models.Question)
@filter_events(lambda instance: instance.type == models.Question.TYPE_CALCULATED_FLOAT)
def remove_calc_dependents(sender, instance, **kwargs):
    update_calc_dependents(
        instance.slug, old_expr=instance.calc_expression, new_expr="false"
    )


# Update calculated answer on post_save
#
# Try to update the calculated answer value whenever a mutation on a possibly
# related model is performed.


@receiver(post_save, sender=models.Question)
@disable_raw
@filter_events(lambda instance: instance.type == models.Question.TYPE_CALCULATED_FLOAT)
def update_calc_from_question(sender, instance, created, update_fields, **kwargs):
    # TODO optimize: only update answers if calc_expression is updated
    # needs to happen during save() or __init__()

    for document in models.Document.objects.filter(form__questions=instance):
        update_or_create_calc_answer(instance, document)


@receiver(post_save, sender=models.FormQuestion)
@disable_raw
@filter_events(
    lambda instance: instance.question.type == models.Question.TYPE_CALCULATED_FLOAT
)
def update_calc_from_form_question(sender, instance, created, **kwargs):
    for document in instance.form.documents.all():
        update_or_create_calc_answer(instance.question, document)


@receiver(post_save, sender=models.Answer)
@disable_raw
@filter_events(lambda instance: instance.document and instance.question.calc_dependents)
def update_calc_from_answer(sender, instance, **kwargs):
    # If there is no document on the answer it means that it's a default
    # answer. They shouldn't trigger a recalculation of a calculated field
    # even when they are technically listed as a dependency.
    # Also skip non-referenced answers.
    if instance.document.family.meta.get("_defer_calculation"):
        return

    for question in models.Question.objects.filter(
        pk__in=instance.question.calc_dependents
    ):
        update_or_create_calc_answer(question, instance.document)


@receiver(post_save, sender=models.Document)
@disable_raw
# We're only interested in table row forms
@filter_events(lambda instance, created: instance.pk != instance.family_id or created)
def update_calc_from_document(sender, instance, created, **kwargs):
    recalculate_answers_from_document(instance)


@receiver(m2m_changed, sender=models.AnswerDocument)
def update_calc_from_answerdocument(sender, instance, **kwargs):
    dependents = instance.document.form.questions.exclude(
        calc_dependents=[]
    ).values_list("calc_dependents", flat=True)

    dependent_questions = list(itertools.chain(*dependents))

    for question in models.Question.objects.filter(pk__in=dependent_questions):
        update_or_create_calc_answer(question, instance.document)



from caluma.caluma_form import models, structure
from caluma.caluma_form.jexl import QuestionJexl


def update_calc_dependents(slug, old_expr, new_expr):
    jexl = QuestionJexl()
    old_q = set(
        list(jexl.extract_referenced_questions(old_expr))
        + list(jexl.extract_referenced_mapby_questions(old_expr))
    )
    new_q = set(
        list(jexl.extract_referenced_questions(new_expr))
        + list(jexl.extract_referenced_mapby_questions(new_expr))
    )

    to_add = new_q - old_q
    to_remove = old_q - new_q

    questions = models.Question.objects.filter(pk__in=list(to_add | to_remove))

    for question in questions:
        deps = set(question.calc_dependents)
        if question.slug in to_add:
            deps.add(slug)
        else:
            deps.remove(slug)
        question.calc_dependents = list(deps)
        question.save()


def update_or_create_calc_answer(question, document):
    root_doc = document.family

    struc = structure.FieldSet(root_doc, root_doc.form)
    field = struc.get_field(question.slug)

    # skip if question doesn't exist in this document structure
    if field is None:
        return

    jexl = QuestionJexl(
        {"form": field.form, "document": field.document, "structure": field.parent()}
    )

    # Ignore errors because we evaluate greedily as soon as possible. At
    # this moment we might be missing some answers or the expression might
    # be invalid, in which case we return None
    value = jexl.evaluate(field.question.calc_expression, raise_on_error=False)

    models.Answer.objects.update_or_create(
        question=question, document=field.document, defaults={"value": value}
    )


def recalculate_answers_from_document(instance):
    if (instance.family or instance).meta.get("_defer_calculation"):
        return
    for question in models.Form.get_all_questions(
        [(instance.family or instance).form_id]
    ).filter(type=models.Question.TYPE_CALCULATED_FLOAT):
        update_or_create_calc_answer(question, instance)
